
# THIS FILE IS GENERATED FROM refellips SETUP.PY
short_version = '0.0.2'
version = '0.0.2'
full_version = '0.0.2'
git_revision = '46e54b6fb70bd47acc48f301c6af6c6af598bee7'
release = True
if not release:
    version = full_version
